"""Configuration file for `logodetect`.

Note: leave this configuration untouched, unless you know what you're doing.
The parameters that can be tuned in the "get_recognizer_config" function below
all concern the
- Detector (first phase)
- Classifier (second phase)
- Data augmentation
- and the devices you want to use (CPU, GPU, etc.)
"""

import os

# Base configuration start
if "LOGOS_RECOGNITION" in os.environ:
    PATH_GLOB = os.environ["LOGOS_RECOGNITION"]
else:
    PATH_GLOB = os.path.expanduser(os.path.join("~", ".hkt", "logodetect"))

PATH_MODELS = os.path.join(PATH_GLOB, "models")
PATH_DATA = os.path.join(PATH_GLOB, "data")
PATH_VIDEO = os.path.join(PATH_DATA, "test_videos")
PATH_IMAGE = os.path.join(PATH_DATA, "test_images")
PATH_EXEMPLARS = os.path.join(PATH_DATA, "exemplars")

SAMPLE_VIDEOS = [
    "test_video_small.mp4",
    "test_video_small_50ms.mp4",
    "football_redbull_small.mp4",
]
SAMPLE_BRANDS = [
    ["pepsi", "redbull", "heineken", "stellaartois"],
    ["pepsi"],
    ["redbull"],
]
CHOICE = 1

VIDEO_FILENAME = os.path.join(PATH_VIDEO, SAMPLE_VIDEOS[CHOICE])
IMAGE_FILENAME = os.path.join(PATH_IMAGE, "test_image_small.png")
BRAND_LOGOS = SAMPLE_BRANDS[CHOICE]

TEST_VIDEO = os.path.join(PATH_VIDEO, "test_video_small_50ms.mp4")
TEST_IMAGE = os.path.join(PATH_IMAGE, "test_image_small.png")
# Base configuration end


def get_recognizer_config(config: dict = None) -> dict:
    """If no config is provided, the default values are chosen,
    otherwise get overwritten.

    :param config: A config dictionary with keys a subset of the below base_config
    :return: the updated (or base) config dictionary
    """

    base_config = {
        "DETECTOR": "detectors.faster_rcnn",
        "DETECTOR_ALG": "binary_fasterrcnn_resnet50",
        "DETECTOR_WEIGHTS": os.path.join(PATH_GLOB, "models", "detector.pth"),
        "USE_CLASSIFIER": True,
        "CLASSIFIER": "knn",  # choose between "knn" and "siamese"
        "CLASSIFIER_WEIGHTS": os.path.join(PATH_MODELS, "classifier_resnet18.pth"),
        # Distance measure used. Choose from: cosine, minkowski_1, or minkowski_2
        "DISTANCE": "cosine",
        # Determined empirically: How far apart are detections and exemplars allowed to be to count as recognized?
        "MAX_DISTANCE": 0.010,
        "PATH_EXEMPLARS_EMBEDDINGS": os.path.join(PATH_DATA, "exemplars_siamese.pkl"),
        "LOAD_EMBEDDINGS": False,
        "EMBEDDING_SIZE": 345,  # Embedding vector length
        # Change this to use your own embedding algorithm
        "EMBEDDER_ALG": "siamese_resnet18",
        "EMBEDDER_WEIGHTS": os.path.join(PATH_MODELS, "embedder.pth"),
        "EMBEDDER_IMG_SIZE": 100,  # Pixel height or width of a square image
        # Which file format do the exemplars have
        "EXEMPLARS_FORMAT": "jpg",
        # Resize exemplars to this size
        "IMAGE_RESIZE": (100, 100),
        # Detection confidence, e.g. 0.9 means that only objects p>=0.9 get detected for the next phase
        "MIN_CONFIDENCE": 0.9,
        # Exemplar augmentation parameters. See `logodetect/augmenters.py` for more examples.
        "AUGMENTER_PARAMS": {
            "Multiply": [0.5, 1.5],  # mu
            "GaussianBlur": [0.4],  # sigma
            "AdditiveGaussianNoise": [0.2 * 255],  # scale
            "AffineShear": [-25, 25],  # shear
            "AffineRotate": [-25, 25],  # rotate
        },
        "DEVICE": "cpu",  # {cpu, cuda:1, cuda:2, ...}
        "EMBEDDER_DEVICE": "cpu",
        "DETECTOR_DEVICE": "cpu",
        "CLASSIFIER_DEVICE": "cpu",
    }

    classifier = base_config.get("CLASSIFIER")
    base_config["CLASSIFIER_ALG"] = (
        "knn" if classifier == "knn" else "binary_stacked_resnet18"
    )

    if config:
        base_config.update(config)
    return base_config



"""Detector module."""

# Standard library:
import os

# Pip packages:
import numpy as np
import torch
from PIL import Image

# Current library:
from logodetect import detectors
from logodetect.utils import image_to_gpu_tensor
from logodetect.constants import get_recognizer_config


class Detector:
    """Detect from input images. Users only need to implement the "predict" function
    below to implement a Detector"""

    def __init__(self, config: dict = None):
        self.config = get_recognizer_config(config)
        model = detectors.get(self.config.get("DETECTOR_ALG"))
        self.model = model(
            self.config.get("DETECTOR_DEVICE"), self.config.get("DETECTOR_WEIGHTS")
        )

    @torch.no_grad()
    def predict(self, image: Image) -> dict:
        """predict detections (boxes, labels, scores and brand) from an input image.

        :param image: PIL.Image
        :return: dictionary with detections
        """
        image = image_to_gpu_tensor(image, self.config.get("DETECTOR_DEVICE"))
        # Always returns list with one element:
        raw_detections = self.model(image)[0]
        return self._process_detections(raw_detections)

    def _process_detections(self, detections: dict) -> dict:
        # Move to cpu:
        detections = self._detections_to_cpu(detections)
        # keep the ones above some (potential) threshold
        selections = np.array(detections["scores"] > self.config.get("MIN_CONFIDENCE"))
        return self._select_detections(detections, selections)

    def _detections_to_cpu(self, detections):
        """Moves all the fields of 'detections' to the CPU."""
        detections["boxes"] = detections["boxes"].cpu().numpy()
        detections["labels"] = detections["labels"].cpu().numpy()
        detections["scores"] = detections["scores"].cpu().numpy()
        return detections

    def _select_detections(self, detections, selections):
        """Filter detections, keep only the selected."""
        detections["boxes"] = detections["boxes"][selections]
        detections["labels"] = detections["labels"][selections]
        detections["scores"] = detections["scores"][selections]
        detections["brands"] = []
        return detections



"""Global utilities."""

# Standard library:
import os

# Pip packages:
import numpy as np
import pandas as pd
from PIL import Image
from torchvision.transforms import functional as F
import torch
from typing import Tuple


def open_and_resize(path: str, image_resize: Tuple[int, int]) -> Image.Image:
    """Checks if image is valid and moves it to the GPU.

    :param path: path to image
    :param image_resize: tuple of two integers
    :return: resized PIL.Image
    """
    return Image.open(path).convert("RGB").resize(image_resize)


def image_to_gpu_tensor(image: Image.Image, device: str) -> torch.Tensor:
    """Checks if image is valid and moves it to the GPU.

    :param image: PIL.Image
    :param device: device to run op on
    :return:
    """
    if isinstance(image, Image.Image):
        image = np.array(image)
    if len(image.shape) != 3 or image.shape[2] != 3:
        raise Exception(
            "'predict' method takes a 3D image as input \
            of shape (H, W, 3). Instead got {}".format(
                image.shape
            )
        )
    return F.to_tensor(image).unsqueeze(0).to(device)


def clean_name(filename: str) -> str:
    """Clean file name

    :param filename: name of the file you want to clean.

    Example:
    >> ' '.join(sorted(set(''.join(list(set(brands))))))
    >> "& ' + - 1 2 3 4 ? a b c d e f g h i j kl m n
        o p q r s t u v w x y z \udcbc \udcc3 \udcfc"
    """
    name, extension = os.path.splitext(os.path.basename(filename))
    brand = name.split("_")[0]
    return brand.encode("ascii", "replace").decode()


def save_df(vectors, file_names, path, net_type="") -> None:
    """Save image vectors and brands stored in file
    names as pandas DataFrame. Only used for visualisation, e.g. in notebooks.

    :param vectors:
    :param file_names:
    :param path:
    :param net_type:
    :return:
    """
    vectors_list = [v for v in vectors]
    brands = [clean_name(n) for n in file_names]
    logos_df = pd.DataFrame({"brand": brands, "img_vec": vectors_list})
    logos_df.to_pickle(path + "{}.pkl".format(net_type))



import os
import glob
from importlib import import_module

import numpy as np
from tqdm import tqdm
import matplotlib.pyplot as plt
import cv2
from moviepy.editor import (
    VideoFileClip,
    VideoClip,
    AudioFileClip,
    ImageClip,
    concatenate_videoclips,
)

from logodetect.utils import clean_name
from logodetect.constants import get_recognizer_config, BRAND_LOGOS


class Recognizer(object):
    """Recognizer

    The Recognizer is the main class for detecting and recognizing logos.
    First, we detect potential overlay boxes in video frames or images. Then,
    if a classifier algorithm has been specified, the detected boxes get labeled
    accordingly and the result is stored as output file.
    """

    def __init__(self, exemplars_path: str = None, config=None) -> None:
        self.config = get_recognizer_config(config)
        self.video = None
        self.audio = None
        self.frame_duration = None
        self.total_frames = None
        if exemplars_path:
            self.set_exemplars(exemplars_path)

        self.detector = import_module(
            f"logodetect.{self.config.get('DETECTOR')}"
        ).Detector(config=self.config)
        if self.config.get("USE_CLASSIFIER"):
            self.classifier = import_module(
                f"logodetect.classifiers.{self.config.get('CLASSIFIER')}"
            ).Classifier(exemplar_paths=self.exemplar_paths, config=self.config)

    def set_exemplars(self, exemplars_path: str) -> None:
        self.exemplars_path = exemplars_path
        all_paths = glob.glob(
            os.path.join(
                self.exemplars_path, "*.{}".format(self.config.get("EXEMPLARS_FORMAT"))
            )
        )
        self.exemplar_paths = [
            path for path in all_paths if clean_name(path) in BRAND_LOGOS
        ]
        self.exemplars_set = sorted(
            set([clean_name(path) for path in self.exemplar_paths])
        )
        self.cmap = plt.cm.get_cmap("jet", len(self.exemplars_set))

    def predict(self, video_filename: str, output_appendix: str = "_output") -> None:
        """Predict recognitions, which will come in the following form:

        recognitions = [{
            'boxes': [] or 2D array (float32),
            'labels': [] or 1D array (int),
            'scores': [] or 1D array (float32),
            'brands': [] or 1D array (str)
            }, {}, {}, ...]

        :param video_filename: file name of the video to process
        :param output_appendix: this string will be appended to the name of the processed video
        :return: None, processed video will be saved
        """
        self.set_video_source(video_filename)

        recognitions = []
        sub_clip_handle = self.video.subclip(0, self.video.end)
        for frame in tqdm(
            sub_clip_handle.iter_frames(),
            total=self.total_frames,
            desc="Processing video",
        ):
            recognitions = self.compute_recognitions(frame, recognitions)
        predicted_video = self.draw_video(recognitions)
        self.save_video(predicted_video, video_filename, output_appendix)

    def predict_image(
        self, image_filename: str, output_appendix: str = "_output"
    ) -> None:
        """Predict recognitions for a single input image.

        :param image_filename: file name of the image to process
        :param output_appendix: this string will be appended to the name of the processed image
        :return: None, processed image will be saved
        """
        image = cv2.imread(image_filename)
        recognitions = self.compute_recognitions(image)
        predicted_image = self.draw_overlay_boxes(image, recognitions[0])
        self.save_image(predicted_image, image_filename, output_appendix)

    def compute_recognitions(
        self, frame: np.ndarray, recognitions: list = None
    ) -> list:
        """Compute new recognitions for this frame and returns the augmented list.

        :param frame: current frame
        :param recognitions: previous list of recognitions
        :return: updated list of recognitions
        """
        if not recognitions:
            recognitions = []
        detections = self.detector.predict(frame)
        if self.config.get("USE_CLASSIFIER"):
            classifications = self.classifier.predict(detections, frame)
            recognitions.append(classifications)
        else:
            recognitions.append(detections)
        return recognitions

    def set_video_source(self, video_filename: str) -> None:
        """Set all video related properties for the specified video

        :param video_filename: file name of the video to process
        :return: None
        """
        self.video = VideoFileClip(video_filename)
        self.audio = AudioFileClip(video_filename)
        self.total_frames = int(self.video.reader.nframes)
        self.frame_duration = 1 / self.video.fps

    def draw_video(self, recognitions: list) -> VideoClip:
        """Draw the recognitions found for the video into the video frames
        and store them in "all_frames".

        :param recognitions: list of recognitions.
        :return:
        """
        all_frames = []
        sub_clip_handle = self.video.subclip(0, self.video.end)
        for idx, frame in enumerate(
            tqdm(
                sub_clip_handle.iter_frames(),
                total=self.total_frames,
                desc="Rendering video",
            )
        ):
            result = self.draw_overlay_boxes(frame, recognitions[idx])
            new_frame = ImageClip(result).set_duration(self.frame_duration)
            all_frames.append(new_frame)
        return concatenate_videoclips(all_frames)

    def draw_overlay_boxes(self, image: np.ndarray, recognition: dict) -> np.ndarray:
        """If there a recognitions for this image, draw overlay
        boxes and text on the image.

        :param image: input image
        :param recognition: recognition object
        :return: processed image with overlay boxes
        """
        if len(recognition["boxes"]) != 0:
            boxes = recognition["boxes"]
            scores = recognition["scores"]
            brands = recognition["brands"]
            font = cv2.FONT_HERSHEY_SIMPLEX

            for box, score, brand in zip(boxes, scores, brands):
                label = self.exemplars_set.index(brand)
                color = tuple(np.array(self.cmap(label))[:3] * 255)
                box = [int(b) for b in box]
                top_left, bottom_right = tuple(box[:2]), tuple(box[2:])
                image = cv2.rectangle(image, top_left, bottom_right, color, 2)
                text = "{}: {:.2f}".format(brand, score)
                cv2.putText(image, text, top_left, font, 0.7, color, 2)

        return image

    @staticmethod
    def save_image(
        image: np.ndarray, image_filename: str, output_appendix: str = "_output"
    ) -> None:
        """Save the resulting image.

        :param image: processed image
        :param image_filename: original file name
        :param output_appendix: appendix to add to file
        :return: None
        """
        output_filename = append_to_file_name(image_filename, output_appendix)
        print(f"Saved resulting image as {output_filename}.")
        cv2.imwrite(output_filename, image)

    def save_video(
        self, video: VideoClip, video_filename: str, output_appendix: str = "_output"
    ) -> None:
        """Save the resulting video.

        :param video: the processed VideoClip
        :param video_filename: original file name
        :param output_appendix: appendix to add to file
        :return: None
        """
        output_filename = append_to_file_name(video_filename, output_appendix)
        video.set_audio(self.audio)
        video.write_videofile(
            output_filename,
            codec="libx264",
            audio_codec="aac",
            temp_audiofile=output_filename + ".tmp",
            remove_temp=True,
            fps=self.video.fps,
        )


def append_to_file_name(video_filename: str, output_appendix: str) -> str:
    name, extension = os.path.splitext(video_filename)
    return video_filename.replace(extension, f"{output_appendix}{extension}")



from setuptools import setup
from setuptools import find_packages

setup(
    name="logodetect",
    version="1.1.3",
    description="One-shot logo detection for videos and images.",
    long_description="One-shot logo detection for videos and images.",
    url="https://github.com/Heldenkombinat/logodetect",
    download_url="https://github.com/Heldenkombinat/logodetect/tarball/1.1.0",
    author="Jorge Davila Chacon",
    author_email="jorge@heldenkombinat.com",
    install_requires=[
        "click>=7.1.1",
        "Cython>=0.29.15",
        "Flask>=1.1.1",
        "Flask-Cors>=3.0.9",
        "gunicorn>=20.0.4",
        "imgaug>=0.4.0",
        "jupyterlab==3.2.4",
        "matplotlib>=2.2.5",
        "moviepy>=1.0.1",
        "numpy>=1.18.2",
        "opencv-python>=4.2.0.32",
        "pandas>=1.0.3",
        "pytest>=6.2.4",
        "scikit-image>=0.16.2",
        "scikit-learn>=0.22.1",
        "scipy>=1.4.1",
        "torch>=1.9.0",
        "torchvision>=0.10.0",
        "tqdm>=4.42.1",
    ],
    extras_require={
        "tests": ["pytest", "pytest-pep8", "pytest-cov", "mock", "pylint", "autopep8"],
        "dev": ["black", "pre-commit", "flask", "gunicorn"],
    },
    entry_points={"console_scripts": ["logodetect=logodetect.cli:cli"]},
    packages=find_packages(),
    license="GNU AGPLv3",
    zip_safe=False,
    classifiers=[
        "Development Status :: 4 - Beta",
        "Intended Audience :: Developers",
        "Environment :: Console",
        "Operating System :: POSIX :: Linux",
        "Programming Language :: Python :: 3",
    ],
)



"""
All "stacked" models will take as input the channel-wise
concatenation of two images and output whether it's the same class or not.
Input should always be exemplar ON TOP of detection, i.e.
    input[:3, :, :] is the detection and input[3:, :, :] is the exemplar
"""

import torch
import torch.nn as nn
import torchvision
from sklearn.neighbors import KNeighborsClassifier

ALL_ARCHITECTURES = [
    "binary_stacked_resnet50",
    "binary_stacked_resnet18",
    "siamese_resnet18",
    "knn",
]


def get(function_name: str):
    """Get a model architecture by function name.

    :param function_name: function name to call
    :return: the function created from the name.
    """
    if function_name not in ALL_ARCHITECTURES:
        raise ValueError(
            f"{function_name} is not a valid model architecture, choose from {ALL_ARCHITECTURES}."
        )
    return globals()[f"_{function_name}"]


def _knn(n_neighbors, metric, *args, **kwargs) -> KNeighborsClassifier:
    return KNeighborsClassifier(n_neighbors=n_neighbors, metric=metric, *args, **kwargs)


def load_model_weights(model, device: str, model_weights: str) -> nn.Sequential:
    checkpoint = torch.load(model_weights, map_location=device)
    model.load_state_dict(checkpoint["model"])
    return nn.Sequential(model, nn.Sigmoid())


def _binary_stacked_resnet50(device: str, model_weights: str):
    """Creates a ResNet50 with a 6-channel input."""
    # Load standard, built-in ResNet 50 architecture:
    model = torchvision.models.resnet50(pretrained=False)

    # Get pre-trained weights of first layer:
    conv1_weight = model.conv1.weight
    # Duplicate array:
    duplicated_weights = torch.cat((conv1_weight, conv1_weight), dim=1)

    # Replace the 3-channel input with 6-channel input:
    model.conv1 = nn.Conv2d(
        6, 64, kernel_size=(7, 7), stride=(2, 2), padding=(3, 3), bias=False
    )

    # Update layer with duplicated pre-trained weights:
    model.conv1.weight = torch.nn.Parameter(data=duplicated_weights)

    # Replace the final layer and add a sigmoid on top:
    model.fc = nn.Linear(model.fc.in_features, out_features=1, bias=True)

    model = load_model_weights(model, device, model_weights)

    return model.eval().to(device)


def _binary_stacked_resnet18(device: str, model_weights: str):
    """Creates a ResNet18 with a 6-channel input."""
    # Load standard architecture:
    model = torchvision.models.resnet18(pretrained=False)
    return _load_binary_stacked_net(model, device, model_weights)


def _load_binary_stacked_net(model, device: str, model_weights: str):
    """Creates a ResNet18 with a 6-channel input."""
    # Replace the 3-channel input with 6-channel input:
    model.conv1 = nn.Conv2d(
        6, 64, kernel_size=(7, 7), stride=(2, 2), padding=(3, 3), bias=False
    )
    # Replace the final layer and add a sigmoid on top:
    model.fc = nn.Linear(model.fc.in_features, out_features=1, bias=True)

    model = load_model_weights(model, device, model_weights)
    return model.eval().to(device)


def _siamese_resnet18(device: str, model_weights: str, model_out: int = 345):
    """Loads a pre-trained ResNet18 for Siamese network."""
    # Load standard architecture:
    model = torchvision.models.resnet18(pretrained=False)

    # Replace the final layer:
    model.fc = nn.Linear(model.fc.in_features, model_out)

    checkpoint = torch.load(model_weights, map_location=device)
    model.load_state_dict(checkpoint["state_dict"])

    return model.eval().to(device)
